@@ -35,14 +35,16 @@
 #endif
 
 #include "bamf-view.h"
+#include "bamf-view-private.h"
 #include "bamf-factory.h"
 #include "bamf-window.h"
+#include "bamf-marshal.h"
 
 #include <dbus/dbus.h>
 #include <dbus/dbus-glib.h>
 #include <dbus/dbus-glib-lowlevel.h>
 
-G_DEFINE_TYPE (BamfView, bamf_view, G_TYPE_OBJECT);
+G_DEFINE_TYPE (BamfView, bamf_view, G_TYPE_INITIALLY_UNOWNED);
 
 #define BAMF_VIEW_GET_PRIVATE(o) \
   (G_TYPE_INSTANCE_GET_PRIVATE ((o), BAMF_TYPE_VIEW, BamfViewPrivate))
@@ -64,7 +66,8 @@ enum
   RUNNING_CHANGED,
   URGENT_CHANGED,
   VISIBLE_CHANGED,
-  
+  NAME_CHANGED,
+ 
   LAST_SIGNAL,
 };
 
@@ -87,9 +90,12 @@ struct _BamfViewPrivate
   DBusGProxy      *proxy;
   gchar           *path;
   gchar           *type;
+  gchar           *local_icon;
+  gchar           *local_name;
   guint            checked_flags;
   guint            set_flags;
   gboolean         is_closed;
+  gboolean         sticky;
 };
 
 static void
@@ -147,6 +153,9 @@ bamf_view_get_children (BamfView *view)
   if (BAMF_VIEW_GET_CLASS (view)->get_children)
     return BAMF_VIEW_GET_CLASS (view)->get_children (view);
 
+  if (!bamf_view_remote_ready (view))
+    return NULL;
+
   priv = view->priv;
 
   if (!dbus_g_proxy_call (priv->proxy,
@@ -187,6 +196,9 @@ bamf_view_get_boolean (BamfView *self, const char *method_name, guint flag)
   if (bamf_view_flag_is_set (self, flag))
     return bamf_view_get_flag (self, flag);
 
+  if (!bamf_view_remote_ready (self))
+    return FALSE;
+
   if (!dbus_g_proxy_call (priv->proxy,
                           method_name,
                           &error,
@@ -255,6 +267,48 @@ bamf_view_is_urgent (BamfView *self)
   return bamf_view_get_boolean (self, "IsUrgent", BAMF_VIEW_URGENT_FLAG);
 }
 
+void
+bamf_view_set_name (BamfView *view, const char *name)
+{
+  g_return_if_fail (BAMF_IS_VIEW (view));
+  
+  if (!g_strcmp0 (name, view->priv->local_name))
+    return;
+
+  view->priv->local_name = g_strdup (name);
+}
+
+void
+bamf_view_set_icon (BamfView *view, const char *icon)
+{
+  g_return_if_fail (BAMF_IS_VIEW (view));
+  
+  view->priv->local_icon = g_strdup (icon);
+}
+
+gboolean 
+bamf_view_is_sticky (BamfView *view)
+{
+  g_return_val_if_fail (BAMF_IS_VIEW (view), FALSE);
+  
+  return view->priv->sticky;
+}
+
+void 
+bamf_view_set_sticky (BamfView *view, gboolean value)
+{
+  g_return_if_fail (BAMF_IS_VIEW (view));
+  
+  if (value == view->priv->sticky)
+    return;
+  
+  if (value)
+    g_object_ref_sink (view);
+  else
+    g_object_unref (view);
+  
+  view->priv->sticky = value;
+}
 
 gchar *
 bamf_view_get_icon (BamfView *self)
@@ -269,6 +323,9 @@ bamf_view_get_icon (BamfView *self)
   if (BAMF_VIEW_GET_CLASS (self)->get_icon)
     return BAMF_VIEW_GET_CLASS (self)->get_icon (self);
 
+  if (!bamf_view_remote_ready (self))
+    return g_strdup (priv->local_icon);
+
   if (!dbus_g_proxy_call (priv->proxy,
                           "Icon",
                           &error,
@@ -297,6 +354,9 @@ bamf_view_get_name (BamfView *self)
   
   if (BAMF_VIEW_GET_CLASS (self)->get_name)
     return BAMF_VIEW_GET_CLASS (self)->get_name (self);
+
+  if (!bamf_view_remote_ready (self))
+    return priv->local_name;
     
   if (!dbus_g_proxy_call (priv->proxy,
                           "Name",
@@ -314,6 +374,12 @@ bamf_view_get_name (BamfView *self)
   return name;
 }
 
+gboolean 
+bamf_view_remote_ready (BamfView *view)
+{
+  return BAMF_IS_VIEW (view) && view->priv->proxy;
+}
+
 const gchar *
 bamf_view_get_view_type (BamfView *self)
 {
@@ -337,7 +403,7 @@ bamf_view_get_view_type (BamfView *self)
                           G_TYPE_STRING, &type,
                           G_TYPE_INVALID))
     {
-      g_warning ("Failed to fetch view type: %s", error->message);
+      g_warning ("Failed to fetch view type at %s: %s", dbus_g_proxy_get_path (priv->proxy), error->message);
       g_error_free (error);
       return NULL;
     }
@@ -346,11 +412,15 @@ bamf_view_get_view_type (BamfView *self)
   return type;
 }
 
-static void
-bamf_view_on_closed (DBusGProxy *proxy, BamfView *self)
+BamfClickBehavior 
+bamf_view_get_click_suggestion (BamfView *self)
 {
-  self->priv->is_closed = TRUE;
-  g_signal_emit (G_OBJECT (self), view_signals[CLOSED], 0);
+  g_return_val_if_fail (BAMF_IS_VIEW (self), BAMF_CLICK_BEHAVIOR_NONE);
+
+  if (BAMF_VIEW_GET_CLASS (self)->click_behavior)
+    return BAMF_VIEW_GET_CLASS (self)->click_behavior (self);
+    
+  return BAMF_CLICK_BEHAVIOR_NONE;
 }
 
 static void
@@ -382,6 +452,17 @@ bamf_view_on_active_changed (DBusGProxy *proxy, gboolean active, BamfView *self)
   g_object_notify (G_OBJECT (self), "active");
 }
 
+static void
+bamf_view_on_name_changed (DBusGProxy*  proxy,
+                           const gchar* old_name,
+                           const gchar* new_name,
+                           BamfView*    self)
+{
+  self->priv->local_name = g_strdup (new_name);
+
+  g_signal_emit (self, view_signals[NAME_CHANGED], 0, old_name, new_name);
+}
+
 static void
 bamf_view_on_running_changed (DBusGProxy *proxy, gboolean running, BamfView *self)
 {
@@ -409,6 +490,69 @@ bamf_view_on_user_visible_changed (DBusGProxy *proxy, gboolean visible, BamfView
   g_object_notify (G_OBJECT (self), "user-visible");
 }
 
+static void
+bamf_view_on_closed (DBusGProxy *proxy, BamfView *self)
+{
+  BamfViewPrivate *priv;
+  
+  priv = self->priv;
+
+  priv->is_closed = TRUE;
+  
+  if (priv->sticky && priv->proxy)
+    {
+      dbus_g_proxy_disconnect_signal (priv->proxy,
+                                      "ActiveChanged",
+                                      (GCallback) bamf_view_on_active_changed,
+                                      self);
+
+      dbus_g_proxy_disconnect_signal (priv->proxy,
+                                      "Closed",
+                                      (GCallback) bamf_view_on_closed,
+                                      self);
+
+      dbus_g_proxy_disconnect_signal (priv->proxy,
+                                      "ChildAdded",
+                                      (GCallback) bamf_view_on_child_added,
+                                      self);
+
+      dbus_g_proxy_disconnect_signal (priv->proxy,
+                                      "ChildRemoved",
+                                      (GCallback) bamf_view_on_child_removed,
+                                      self);
+
+      dbus_g_proxy_disconnect_signal (priv->proxy,
+                                      "RunningChanged",
+                                      (GCallback) bamf_view_on_running_changed,
+                                      self);
+
+      dbus_g_proxy_disconnect_signal (priv->proxy,
+                                     "UrgentChanged",
+                                     (GCallback) bamf_view_on_urgent_changed,
+                                     self);
+  
+      dbus_g_proxy_disconnect_signal (priv->proxy,
+                                     "UserVisibleChanged",
+                                     (GCallback) bamf_view_on_user_visible_changed,
+                                     self);
+      g_object_unref (priv->proxy);
+      priv->proxy = NULL;
+    }
+  
+  g_object_ref (self);
+  
+  // must be emitted before path is cleared as path is utilized in cleanup
+  g_signal_emit (G_OBJECT (self), view_signals[CLOSED], 0);
+
+  if (priv->path)
+    {
+      g_free (priv->path);
+      priv->path = NULL;
+    }
+    
+  g_object_unref (self);
+}
+
 static void
 bamf_view_set_property (GObject *object, guint property_id, const GValue *value, GParamSpec *pspec)
 {
@@ -418,11 +562,6 @@ bamf_view_set_property (GObject *object, guint property_id, const GValue *value,
 
   switch (property_id)
     {
-      case PROP_PATH:
-        self->priv->path = g_value_dup_string (value);
-
-        break;
-        
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
     }
@@ -521,30 +660,41 @@ bamf_view_dispose (GObject *object)
                                      "UserVisibleChanged",
                                      (GCallback) bamf_view_on_user_visible_changed,
                                      view);
+
+      dbus_g_proxy_disconnect_signal (priv->proxy,
+                                     "NameChanged",
+                                     (GCallback) bamf_view_on_name_changed,
+                                     view);
+
       g_object_unref (priv->proxy);
       priv->proxy = NULL;
     }
 }
 
-static void
-bamf_view_constructed (GObject *object)
+const char * 
+bamf_view_get_path (BamfView *view)
+{
+  g_return_val_if_fail (BAMF_IS_VIEW (view), NULL);
+  
+  return view->priv->path;
+}
+
+void
+bamf_view_set_path (BamfView *view, const char *path)
 {
-  BamfView *view;
   BamfViewPrivate *priv;
   
-  if (G_OBJECT_CLASS (bamf_view_parent_class)->constructed)
-    G_OBJECT_CLASS (bamf_view_parent_class)->constructed (object);
-
-  view = BAMF_VIEW (object);
-
+  g_return_if_fail (BAMF_IS_VIEW (view));
+  
   priv = view->priv;
+  priv->is_closed = FALSE;
 
-  if (priv->proxy)
+  if (priv->path)
     {
-      g_warning ("bamf_view_set_path called multiple times\n");
-      return;
+      g_free (priv->path);
     }
   
+  priv->path = g_strdup (path);
   priv->proxy = dbus_g_proxy_new_for_name (priv->connection,
                                            "org.ayatana.bamf",
                                            priv->path,
@@ -589,6 +739,12 @@ bamf_view_constructed (GObject *object)
                            G_TYPE_BOOLEAN, 
                            G_TYPE_INVALID);
 
+  dbus_g_proxy_add_signal (priv->proxy,
+                           "NameChanged",
+                           G_TYPE_STRING,
+                           G_TYPE_STRING,
+                           G_TYPE_INVALID);
+
   dbus_g_proxy_connect_signal (priv->proxy,
                                "ActiveChanged",
                                (GCallback) bamf_view_on_active_changed,
@@ -630,6 +786,50 @@ bamf_view_constructed (GObject *object)
                                (GCallback) bamf_view_on_user_visible_changed,
                                view,
                                NULL);
+
+  dbus_g_proxy_connect_signal (priv->proxy,
+                               "NameChanged",
+                               (GCallback) bamf_view_on_name_changed,
+                               view,
+                               NULL);
+
+  if (bamf_view_is_sticky (view))
+    {
+      priv->checked_flags = 0x0;
+      
+      if (bamf_view_user_visible (view))
+        {
+          g_signal_emit (G_OBJECT(view), view_signals[VISIBLE_CHANGED], 0, TRUE);
+          g_object_notify (G_OBJECT (view), "user-visible");
+        }
+      
+      if (bamf_view_is_active (view))
+        {
+          g_signal_emit (G_OBJECT(view), view_signals[ACTIVE_CHANGED], 0, TRUE);
+          g_object_notify (G_OBJECT (view), "active");
+        }
+      
+      if (bamf_view_is_running (view))
+        {
+          g_signal_emit (G_OBJECT(view), view_signals[RUNNING_CHANGED], 0, TRUE);
+          g_object_notify (G_OBJECT (view), "running");
+        }
+        
+      if (bamf_view_is_urgent (view))
+        {
+          g_signal_emit (G_OBJECT(view), view_signals[URGENT_CHANGED], 0, TRUE);
+          g_object_notify (G_OBJECT (view), "urgent");
+        }
+    }
+  if (BAMF_VIEW_GET_CLASS (view)->set_path)
+    BAMF_VIEW_GET_CLASS (view)->set_path (view, path);
+}
+
+static void
+bamf_view_constructed (GObject *object)
+{
+  if (G_OBJECT_CLASS (bamf_view_parent_class)->constructed)
+    G_OBJECT_CLASS (bamf_view_parent_class)->constructed (object);
 }
 
 static void
@@ -643,7 +843,7 @@ bamf_view_class_init (BamfViewClass *klass)
   obj_class->get_property = bamf_view_get_property;
   obj_class->set_property = bamf_view_set_property;
 
-  pspec = g_param_spec_string ("path", "path", "path", NULL, G_PARAM_READWRITE | G_PARAM_CONSTRUCT);
+  pspec = g_param_spec_string ("path", "path", "path", NULL, G_PARAM_READABLE);
   g_object_class_install_property (obj_class, PROP_PATH, pspec);
   
   pspec = g_param_spec_boolean ("active", "active", "active", FALSE, G_PARAM_READABLE);
@@ -728,6 +928,16 @@ bamf_view_class_init (BamfViewClass *klass)
   	              g_cclosure_marshal_VOID__BOOLEAN,
   	              G_TYPE_NONE, 1, 
   	              G_TYPE_BOOLEAN);
+
+  view_signals [NAME_CHANGED] =
+        g_signal_new ("name-changed",
+                      G_OBJECT_CLASS_TYPE (klass),
+                      0,
+                      0, NULL, NULL,
+                      bamf_marshal_VOID__STRING_STRING,
+  	              G_TYPE_NONE, 2,
+  	              G_TYPE_STRING,
+                      G_TYPE_STRING);
 }
 
 
@@ -739,6 +949,8 @@ bamf_view_init (BamfView *self)
 
   priv = self->priv = BAMF_VIEW_GET_PRIVATE (self);
 
+  priv->is_closed = TRUE;
+
   priv->connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
   if (priv->connection == NULL)
     {
