@@ -48,6 +48,7 @@ struct _BamfApplicationPrivate
   GList * desktop_file_list;
   char * app_type;
   char * icon;
+  char * wmclass;
   gboolean is_tab_container;
   gboolean show_stubs;
 };
@@ -83,6 +84,19 @@ bamf_application_get_desktop_file (BamfApplication *application)
   return result;
 }
 
+char *
+bamf_application_get_wmclass (BamfApplication *application)
+{
+  BamfApplicationPrivate *priv;
+  char *result = NULL;
+
+  g_return_val_if_fail (BAMF_IS_APPLICATION (application), NULL);
+  priv = application->priv;
+
+  result = g_strdup (priv->wmclass);
+  return result;
+}
+
 static gboolean
 icon_name_is_valid (char *name)
 {
@@ -106,6 +120,7 @@ bamf_application_setup_icon_and_name (BamfApplication *self)
   GList *children, *l;
   const char *class;
   char *icon = NULL, *name = NULL;
+  GError *error;
 
   g_return_if_fail (BAMF_IS_APPLICATION (self));
 
@@ -138,6 +153,25 @@ bamf_application_setup_icon_and_name (BamfApplication *self)
            off the stub menus. */
         self->priv->show_stubs = g_key_file_get_boolean(keyfile, G_KEY_FILE_DESKTOP_GROUP, STUB_KEY, NULL);
       }
+      
+      if (g_key_file_has_key(keyfile, G_KEY_FILE_DESKTOP_GROUP, "X-GNOME-FullName", NULL))
+    	  {
+    	    /* Grab the better name if its available */
+    	    gchar *fullname = NULL;
+    	    error = NULL; 
+    	    fullname = g_key_file_get_locale_string (keyfile, G_KEY_FILE_DESKTOP_GROUP, "X-GNOME-FullName", NULL, &error);
+    	    if (error != NULL)
+    	      {
+    	        g_error_free (error);
+    	        if (fullname)
+    	          g_free (fullname);
+    	      }
+    	    else
+    	      {
+    	        g_free (name);
+    	        name = fullname;
+    	      }
+    	  }
 
       g_object_unref (desktop);
       g_key_file_free(keyfile);
@@ -171,7 +205,7 @@ bamf_application_setup_icon_and_name (BamfApplication *self)
             }
           while (FALSE);
 
-          name = g_strdup (bamf_legacy_window_get_class_name (bamf_window_get_window (window)));
+          name = g_strdup (bamf_legacy_window_get_name (bamf_window_get_window (window)));
         }
         
       if (!icon)
@@ -212,6 +246,18 @@ bamf_application_set_desktop_file (BamfApplication *application,
   bamf_application_setup_icon_and_name (application);
 }
 
+void
+bamf_application_set_wmclass (BamfApplication *application,
+                            char *wmclass)
+{
+  g_return_if_fail (BAMF_IS_APPLICATION (application));
+
+  if (wmclass && wmclass[0] != '\0')
+    application->priv->wmclass = g_strdup (wmclass);
+  else
+    application->priv->wmclass = NULL;
+}
+
 GArray *
 bamf_application_get_xids (BamfApplication *application)
 {
@@ -305,6 +351,34 @@ bamf_application_get_view_type (BamfView *view)
   return g_strdup ("application");
 }
 
+static char *
+bamf_application_get_stable_bus_name (BamfView *view)
+{
+  BamfApplication *self;
+  GList *children, *l;
+  BamfView *child;
+
+  g_return_val_if_fail (BAMF_IS_APPLICATION (view), NULL);
+  self = BAMF_APPLICATION (view);
+
+  if (self->priv->desktop_file)
+    return g_strdup_printf ("application%i", abs (g_str_hash (self->priv->desktop_file)));
+
+  children = bamf_view_get_children (BAMF_VIEW (self));
+  for (l = children; l; l = l->next)
+    {
+      child = l->data;
+
+      if (!BAMF_IS_WINDOW (child))
+        continue;
+
+      return g_strdup_printf ("application%s", 
+                              bamf_legacy_window_get_class_name (bamf_window_get_window (BAMF_WINDOW (child))));
+    }
+  
+  return g_strdup_printf ("application%p", view);
+}
+
 static void
 bamf_application_ensure_flags (BamfApplication *self)
 {
@@ -384,13 +458,6 @@ bamf_application_child_added (BamfView *view, BamfView *child)
                           (GCallback) view_exported, view);
     }
 
-  // If we're not a real application, get some properties from our first child
-  if (application->priv->desktop_file == NULL)
-    {
-      if (bamf_view_get_name (view) == NULL)
-        bamf_view_set_name (view, bamf_view_get_name (child));
-    }
-
   g_signal_connect (G_OBJECT (child), "active-changed",
                     (GCallback) view_active_changed, view);
   g_signal_connect (G_OBJECT (child), "urgent-changed",
@@ -534,6 +601,12 @@ bamf_application_dispose (GObject *object)
       g_free (priv->app_type);
       priv->app_type = NULL;
     }
+
+  if (priv->wmclass)
+    {
+      g_free (priv->wmclass);
+      priv->wmclass = NULL;
+    }
   
   g_signal_handlers_disconnect_by_func (G_OBJECT (bamf_matcher_get_default ()), matcher_favorites_changed, object);
 
@@ -549,6 +622,7 @@ bamf_application_init (BamfApplication * self)
   priv->is_tab_container = FALSE;
   priv->app_type = g_strdup ("system");
   priv->show_stubs = TRUE;
+  priv->wmclass = NULL;
   
   g_signal_connect (G_OBJECT (bamf_matcher_get_default ()), "favorites-changed", 
                     (GCallback) matcher_favorites_changed, self);
@@ -566,6 +640,7 @@ bamf_application_class_init (BamfApplicationClass * klass)
   view_class->child_added = bamf_application_child_added;
   view_class->child_removed = bamf_application_child_removed;
   view_class->get_icon = bamf_application_get_icon;
+  view_class->stable_bus_name = bamf_application_get_stable_bus_name;
 
   g_type_class_add_private (klass, sizeof (BamfApplicationPrivate));
 
@@ -622,6 +697,17 @@ bamf_application_new_from_desktop_files (GList *desktop_files)
   return application;  
 }
 
+BamfApplication *
+bamf_application_new_with_wmclass (char *wmclass)
+{
+  BamfApplication *application;
+  application = (BamfApplication *) g_object_new (BAMF_TYPE_APPLICATION, NULL);
+
+  bamf_application_set_wmclass (application, wmclass);
+
+  return application;
+}
+
 /**
     bamf_application_get_show_stubs:
     @application: Application to check for menu stubs
